#!/usr/bin/perl

use strict;
use FileHandle;
use POSIX ":sys_wait_h";

# Where is the config file and where are all the VCB programs?
my $confname = $ARGV[0] || 'c:\users\administrator.ecs2000\documents\VCBBackup.conf';
my $VCBpath = '"C:\Program Files (x86)\VMware\VMware Consolidated Backup Framework"';

# What is the root directory under which all backups should go?
my($VCBdest, $VCBdestUnix, $TodayDir, $TodayDirUnix);

# Do not edit the next line, the substitutions are done automatically
my $VCBauth = '-h HOST -u USER -p PASS';
use constant MAX_KIDS => 3;

# Put them in directories named after the day we started the backup cycle.
my $day = ('Sun','Mon','Tue','Wed','Thu','Fri','Sat')[(localtime)[6]];

$confname =~ s/^"(.*)"$/$1/; # Delete any quotes to protect spaces in the path

$SIG{CHLD} = 'IGNORE';

my $conffh = new FileHandle;
my $linenum = 0;
my(@RulePattern, @RuleAction, $line, $value);
$conffh->open($confname) or die "Cannot read config file $confname, $!\n";
while(defined ($line = <$conffh>)) {
  $linenum++;
  chomp $line;
  $line =~ s/#.*$//g; # Remove comments
  $line =~ s/^\s+//; # Trim leading and
  $line =~ s/\s+$//; # trailing spaces and
  next if $line =~ /^$/; # skip blank lines
  # Include or exclude a regexp of hosts
  if ($line =~ /^(INCLUDE|EXCLUDE)\s+(.*)$/i) {
    push @RuleAction, uc($1);
    my $pattern = $2;
    $pattern =~ s/\*/.*/g; # Convert any * to .* to make it a regexp
    $pattern = '^' . $pattern . '$'; # Anchor to start and end
    push @RulePattern, $pattern;
  # The VMware server hostname
  } elsif ($line =~ /^HOST\s+(.*)$/i) {
    $value = $1;
    $VCBauth =~ s/HOST/$value/;
  # The VMware server username
  } elsif ($line =~ /^USERNAME\s+(.*)$/i) {
    $value = $1;
    $VCBauth =~ s/USER/$value/;
  # The VMware server password
  } elsif ($line =~ /^PASSWORD\s+(.*)$/i) {
    $value = $1;
    $VCBauth =~ s/PASS/$value/;
  } elsif ($line =~ /^DIRECTORY\s+(.*)$/i) {
    $value = $1;
    $VCBdest = $value;
    $VCBdestUnix = $value;
    $VCBdestUnix =~ s/\\/\//g; # Replace all \ with /
  } else {
    print STDERR "Skipping line $linenum as it does not start with INCLUDE, EXCLUDE, HOST, USERNAME or PASSWORD.\n";
  }
}
$conffh->close;


#
# Get the names and details of every possible VM on this server
#
my ($moref, $name, $uuid, %name2moref, %name2uuid, %uuid2name);
my $vmname = new FileHandle;
#print "$VCBpath\\vcbvmname $VCBauth -s any:\n";
$vmname->open("$VCBpath\\vcbvmname $VCBauth -s any:|") or die "vcbvmname: $!\n";
while(<$vmname>) {
  chomp;
  die "Wrong root password!\n" if /Invalid user/;
  if (/^Found VM/) {
    $moref = undef;
    $name  = undef;
    $uuid  = undef;
  } elsif (/^moref:(\d+)/) {
    # Numeric id of the VM, changes with each ESX server reboot
    $moref = $1;
  } elsif (/^name:(.*)/) {
    # Name of the VM, may contain spaces, set by the VMware administrator
    $name = $1;
  } elsif (/^uuid:([0-9a-fA-F-]+)/) {
    # GUID of the VM, set permanently for the VM
    $uuid = lc($1);
  }
  if (defined $name && defined $moref) {
    $name2moref{$name} = $moref;
  }
  if (defined $name && defined $uuid) {
    $name2uuid{$name} = $uuid;
    $uuid2name{$uuid} = $name;
  }
}
$vmname->close;

## Dump maps from name to moref and name to uuid
#foreach $name (sort keys %name2uuid) {
#  print $name2moref{$name} . "\t" . $name2uuid{$name} . " $name\n";
#}

# By default no VMs are backed up.
my %BackupVM;

# Now go through each VM in turn, looking for a matching rule.
# This gives you a hash of what VM names to backup and what to skip.
print "\n";
foreach $name (sort keys %name2uuid) {
  # Go through the ruleset finding the first match
  my $foundVM = 0;
  foreach my $rulenum (0..$#RulePattern) {
    my $regexp = $RulePattern[$rulenum];
    next unless $name =~ /$regexp/;
    $BackupVM{$name} = 1 if $RuleAction[$rulenum] eq 'INCLUDE';
    $BackupVM{$name} = 0 if $RuleAction[$rulenum] eq 'EXCLUDE';
    $foundVM = 1;
    last;
  }
  # Did we find a regexp match?
  if (! $foundVM) {
    # No, so skip it by default
    print "No match for \"$name\" found, skipping\n";
  } else {
    # We found a matching VM name
    print "Backing up $name\n" if $BackupVM{$name};
    print "Skipping $name\n" unless $BackupVM{$name};
  }
}
print "\n";

#
# Now we have name --> moref and name --> uuid mappings of the VM list
#

# Where are we going to put the backups?
$TodayDir = "$VCBdest\\$day";
$TodayDirUnix = "$VCBdestUnix/$day";
mkdir $TodayDirUnix;

my $now = localtime;

# Now produce 3 lists of target VMs, one list for each thread.
my(%uuids, @uuidlist, @uuidlist1, @uuidlist2, @uuidlist3, $vm);
foreach $vm (sort keys %BackupVM) {
  push @uuidlist, $name2uuid{$vm} if $BackupVM{$vm} && $name2uuid{$vm};
}
@uuidlist1 = splice @uuidlist, 0, ((@uuidlist+1)/3);
@uuidlist2 = splice @uuidlist, 0, ((@uuidlist+1)/2);
@uuidlist3 = @uuidlist;
$uuids{1}  = \@uuidlist1;
$uuids{2}  = \@uuidlist2;
$uuids{3}  = \@uuidlist3;

my (%children, $quit, $childnum) = ((), 0, 0);

while(!$quit) {
  reap(\%children);

  if ((keys %children) < MAX_KIDS) {
    $childnum++;
    print "\n[$$] Child $childnum starting\n";
    if (my $pid = fork()) {
      $children{$pid} = 1;
    } else {
      sleep 1;
      do_child($childnum);
      exit;
    }
  } else {
    $quit = 1;
  }

  sleep 1;
}

# Reap existing kids
while (keys %children) {
  reap(\%children);
  sleep 1;
}

sub do_child {
  my ($childnumber) = @_;

  my ($uuid);
  
  print "In child $childnumber\n";
  foreach $uuid (@{$uuids{$childnumber}}) {
    sleep 2;
    print "\nChild $childnumber starting $uuid2name{$uuid}\n";
    VCBmount($uuid, $TodayDir, $TodayDirUnix);
    print "\nChild $childnumber finished $uuid2name{$uuid}\n";
  }
}

sub reap {
  my ($kids) = @_;
  for (keys %{$kids}) {
    next if waitpid($_, WNOHANG()) != -1;
    delete $kids->{$_};
  }
}



$now = localtime;
print "\n\nFinished parent process at $now\n";

exit 0;

# Backup the given UUID virtual machine to a directory named the same as the VM
# directly under the $rootdir directory. Need forward and backslash versions.
sub VCBmount {
  my($uuid, $rootdir, $rootdirUnix) = @_;

  #
  # We need to delete any old leftover backup snapshots
  #
  # Find the id of the snapshot
  print "Checking for old backup snapshots...\n";
  my $snapshotfh = new FileHandle;
  print "$VCBpath\\vcbsnapshot $VCBauth -f moref:" .
                        $name2moref{$uuid2name{$uuid}} .
                        " _VCB-BACKUP_\n";
  if ($snapshotfh->open("$VCBpath\\vcbsnapshot $VCBauth -f moref:" .
                        $name2moref{$uuid2name{$uuid}} .
                        " _VCB-BACKUP_ |")) {
    my $snapshotid = "";
    while(<$snapshotfh>) {
      chomp;
      if (/^ssid:(.*)\s*$/i) {
        # Found the snapshot id
        $snapshotid = $1;
        print "Found old backup snapshot \"$snapshotid\" in \"" . $uuid2name{$uuid} .
              "\"\n";
      }
    }
    $snapshotfh->close;
    if ($snapshotid =~ /\S/) {
      # There was a snapshot id, so delete the snapshot
      print "Deleting old backup snapshot \"$snapshotid\" from \"" . $uuid2name{$uuid} .
            "\"\n";
      system("$VCBpath\\vcbsnapshot $VCBauth -d moref:" .
             $name2moref{$uuid2name{$uuid}} .
             " ssid:$snapshotid");
    } else {
      print "No backup snapshot found, goodo!\n";
    }
  } else {
    print "VCBsnapshot failed :-(\n";
  }

  # What sub-directory should we put it in
  my $name = $uuid2name{$uuid};
  $name =~ s/[^a-zA-Z0-9_.-]//g; # Clean out nasty characters
  # Delete the old directory and its contents
  if (chdir $rootdirUnix) {
    unlink <$name/*>;
    rmdir $name;
  }

  # Try to backup the filesystem 10 times until the filesystem can be quiesced.
  my $mounter = new FileHandle;
  my $tryagain = 0;
  my $attempts = 0;
  do {
    print "$name: $VCBpath\\vcbmounter $VCBauth -a uuid:$uuid -r $rootdir\\$name -t fullvm -m nbd\n";
    if ($mounter->open("$VCBpath\\vcbmounter $VCBauth -a uuid:$uuid -r $rootdir\\$name -t fullvm -m nbd |")) {
      while(<$mounter>) {
        chomp;
        # Don't print the output if it's boring and irrelevant
        print "$name: $_\n" unless /host certificate|SSLVerify|system libcrypto/;
        if (/Could not quiesce file system/i) {
          print "Could not quiesce file system, trying again\n";
          $tryagain = 1;
          last;
        }
      }
    } else {
      print "Could not backup $name, vcbmounter failed.\n";
    }
    $attempts ++;
  } while ($tryagain && $attempts<10);

  print "Failed after 10 attempts.\n" if $attempts>=10;
  my $now = localtime;
  print "\nFinished $name at $now\n";

  return 1;
}

